/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
/// Java is a trademark of Sun Microsystems, Inc.
package org.xmloperator.java.syntax;

import java.util.Collections;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.Stack;
import java.io.IOException;

import org.xmloperator.java.syntax.JavaToken;
import org.xmloperator.java.syntax.JavaTokenUtil;
import org.xmloperator.java.syntax.JavaTokenImpl;
import org.xmloperator.java.syntax.IllegalSyntaxException;
import org.xmloperator.java.syntax.Localizer;
import org.xmloperator.java.syntax.JavaTokenizer;
import org.xmloperator.java.syntax.CompilationUnitHandler;
import org.xmloperator.java.syntax.IdentifierList;
import org.xmloperator.java.syntax.IdentifierListMap;

/**
 * Java compilation unit parser.
 * Its purpose is identifying the outer classes and their dependencies.
 *
 * Constraints :
 * - Class and interface identifiers should begin with an upper case.
 * - Package, variable and method identifiers should begin with a lower case.
 *
 * If these constraints are not respected then some depencencies could be
 *   missed.
 */
public final class CompilationUnitParser {
  private JavaTokenizer javaTokenizer;
  private CompilationUnitHandler compilationUnitHandler;
  private IdentifierListMap
      fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList;
  private Set primaryPackageIdentifierSet;
  private boolean isWarningPrinted;
  private JavaToken currentJavaToken = null;
  private String packageQualifiedIdentifier = null;
  private final Stack javaTokenStack = new Stack();
  private final Stack classOrInterfaceIdentifierStack = new Stack();
  private final Map
      fromClassOrInterfaceIdentifierToClassOrInterfaceQualifiedIdentifier =
      new HashMap();
  private final JavaToken savedJavaToken = new JavaTokenImpl();

  private static final JavaTokenImpl TOKEN_DUMMY_IDENTIFIER =
      new JavaTokenImpl();
  private static final JavaTokenImpl TOKEN_DOT = new JavaTokenImpl();
  private static final JavaTokenImpl TOKEN_START_SQUARE_BRACKET =
      new JavaTokenImpl();
  private static final JavaTokenImpl TOKEN_SYNCHRONIZED =
      new JavaTokenImpl();

  static {
    TOKEN_DUMMY_IDENTIFIER.setIdentifier(new char[] {'a'}, 0, 1);
    TOKEN_DOT.setSeparator(JavaToken.SEPARATOR_DOT);
    TOKEN_START_SQUARE_BRACKET.setSeparator(
        JavaToken.SEPARATOR_START_SQUARE_BRACKET);
    TOKEN_SYNCHRONIZED.setKeyword(JavaToken.KEYWORD_SYNCHRONIZED);
  }

  /**
   * Parse the tokenized java source and produce compilation unit handled
   *   nofications.
   */
  public static final void parse(
      JavaTokenizer javaTokenizer,
      CompilationUnitHandler compilationUnitHandler,
      IdentifierListMap
      fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList,
      Set primaryPackageIdentifierSet, boolean isWarningPrinted)
      throws IllegalSyntaxException, IOException {
    new CompilationUnitParser(
        javaTokenizer, compilationUnitHandler,
        fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList,
        primaryPackageIdentifierSet, isWarningPrinted).parse();
  }

  private CompilationUnitParser(
      JavaTokenizer javaTokenizer,
      CompilationUnitHandler compilationUnitHandler,
      IdentifierListMap
      fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList,
      Set primaryPackageIdentifierSet, boolean isWarningPrinted) {
    if (javaTokenizer == null || compilationUnitHandler == null)
      throw new IllegalArgumentException();
    this.javaTokenizer = javaTokenizer;
    this.compilationUnitHandler = compilationUnitHandler;
    this.fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList =
        fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList;
    this.primaryPackageIdentifierSet = primaryPackageIdentifierSet;
    this.isWarningPrinted = isWarningPrinted;
  }

  ///////////////
  // Root parsing
  ///////////////

  private void parse() throws IllegalSyntaxException, IOException {
    this.nextToken();
    this.parseCompilationUnit();
  }

  private void parseCompilationUnit()
      throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_PACKAGE) {
      this.nextToken();
      this.packageQualifiedIdentifier = this.parseQualifiedIdentifier();
      this.verifyPackageQualifiedIdentifier(this.packageQualifiedIdentifier);
      this.skipSemiColon();
    }
    this.compilationUnitHandler.startCompilationUnit(
        this.packageQualifiedIdentifier);
    while (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_IMPORT) {
      this.parseImportDeclaration();
    }
    while (this.currentJavaToken.getType() != JavaToken.NO_TOKEN) {
      if (this.currentJavaToken.getSeparator()
          == JavaToken.SEPARATOR_SEMI_COLON)
        this.nextToken();
      else
        this.parseClassOrInterfaceDeclaration();
    }
    this.compilationUnitHandler.endCompilationUnit();
  }

  private void parseImportDeclaration()
      throws IllegalSyntaxException, IOException {
    this.nextToken();
    boolean isMultipleImport = false;
    StringBuffer sb = new StringBuffer();
    while (true) {
      if (this.currentJavaToken.getType() != JavaToken.TYPE_IDENTIFIER)
        this.syntaxError(ERROR_IDENTIFIER_EXPECTED);
      this.currentJavaToken.appendValueTo(sb);
      this.nextToken();
      if (this.currentJavaToken.getSeparator() != JavaToken.SEPARATOR_DOT)
        break;
      this.nextToken();
      if (this.currentJavaToken.getOperator() == JavaToken.OPERATOR_MULTIPLY) {
        this.nextToken();
        isMultipleImport = true;
        break;
      }
      sb.append('.');
    }
    String qualifiedIdentifier = sb.toString();
    if (isMultipleImport) {
      this.verifyPackageQualifiedIdentifier(qualifiedIdentifier);
      IdentifierList classAndInterfaceIdentifierList =
          this.fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList != null ?
          this.fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList.
          getIdentifierList(qualifiedIdentifier) : null;
      if (classAndInterfaceIdentifierList != null)
        for (int k = 0;
             k < classAndInterfaceIdentifierList.getIdentifierCount(); k++) {
          String classOrInterfaceIdentifier =
              classAndInterfaceIdentifierList.getIdentifier(k);
          this.fromClassOrInterfaceIdentifierToClassOrInterfaceQualifiedIdentifier.
              put(classOrInterfaceIdentifier,
                  qualifiedIdentifier + "." + classOrInterfaceIdentifier);
        }
      else if (
          this.fromPackageQualifiedIdentifierToClassAndInterfaceIdentifierList != null)
        this.warning(WARNING_UNKNOWN_PACKAGE + " : " + qualifiedIdentifier);
    }
    else {
      this.verifyImportQualifiedIdentifier(qualifiedIdentifier);
      this.fromClassOrInterfaceIdentifierToClassOrInterfaceQualifiedIdentifier.
          put(getLastIdentifier(qualifiedIdentifier), qualifiedIdentifier);
    }
    this.skipSemiColon();
  }

  //////////////////////////////
  // Class and interface parsing
  //////////////////////////////

  private void parseClassOrInterfaceDeclaration()
      throws IllegalSyntaxException, IOException {
    this.skipModifiers();
    switch (this.currentJavaToken.getKeyword()) {
      case JavaToken.KEYWORD_CLASS:
        this.parseClassDeclaration();
        break;
      case JavaToken.KEYWORD_INTERFACE:
        this.parseInterfaceDeclaration();
        break;
      default:
        this.syntaxError(ERROR_CLASS_OR_INTERFACE_EXPECTED);
    }
  }

  private void parseClassDeclaration()
      throws IllegalSyntaxException, IOException {
    this.nextToken();
    boolean isOuterClass = this.classOrInterfaceIdentifierStack.isEmpty();
    this.classOrInterfaceIdentifierStack.push(this.parseIdentifier());
    this.verifyCurrentClassOrInterfaceIdentifier();
    String extendedClassQualifiedIdentifier = null;
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_EXTENDS) {
      this.nextToken();
      extendedClassQualifiedIdentifier = this.parseType(HANDLING_NO);
    }
    List implementedInterfaceQualifiedIdentifierList =
        Collections.EMPTY_LIST;
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_IMPLEMENTS) {
      this.nextToken();
      implementedInterfaceQualifiedIdentifierList =
          this.parseTypeList(HANDLING_NO);
    }
    if (isOuterClass)
      this.compilationUnitHandler.startClass(
          (String)this.classOrInterfaceIdentifierStack.peek(),
          extendedClassQualifiedIdentifier,
          implementedInterfaceQualifiedIdentifierList);
    this.parseClassOrInterfaceBody();
    if (isOuterClass)
      this.compilationUnitHandler.endClassOrInterface();
    this.classOrInterfaceIdentifierStack.pop();
  }

  private void parseInterfaceDeclaration()
      throws IllegalSyntaxException, IOException {
    this.nextToken();
    boolean isOuterInterface = this.classOrInterfaceIdentifierStack.isEmpty();
    this.classOrInterfaceIdentifierStack.push(this.parseIdentifier());
    this.verifyCurrentClassOrInterfaceIdentifier();
    List extendedInterfaceQualifiedIdentifierList =
        Collections.EMPTY_LIST;
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_EXTENDS) {
      this.nextToken();
      extendedInterfaceQualifiedIdentifierList =
          this.parseTypeList(HANDLING_NO);
    }
    if (isOuterInterface)
      this.compilationUnitHandler.startInterface(
          (String)this.classOrInterfaceIdentifierStack.peek(),
          extendedInterfaceQualifiedIdentifierList);
    this.parseClassOrInterfaceBody();
    if (isOuterInterface)
      this.compilationUnitHandler.endClassOrInterface();
    this.classOrInterfaceIdentifierStack.pop();
  }

  private void parseClassOrInterfaceBody()
      throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getSeparator()
        != JavaToken.SEPARATOR_START_BRACE_BRACKET)
      this.syntaxErrorCharExpected('{');
    this.nextToken();
    while (this.currentJavaToken.getSeparator()
           != JavaToken.SEPARATOR_END_BRACE_BRACKET) {
      if (this.currentJavaToken.getSeparator()
          == JavaToken.SEPARATOR_SEMI_COLON) {
        this.nextToken();
        continue;
      }
      this.skipModifiers();
      if (this.currentJavaToken.getSeparator()
          == JavaToken.SEPARATOR_START_BRACE_BRACKET)
        this.parseBlock();
      else
        this.parseMemberDeclaration();
    }
    this.nextToken();
  }

  private void parseBlock() throws IllegalSyntaxException, IOException {
    this.skipSeparator(JavaToken.SEPARATOR_START_BRACE_BRACKET);
    while (this.currentJavaToken.getSeparator()
           != JavaToken.SEPARATOR_END_BRACE_BRACKET)
      this.parseBlockStatement();
    this.nextToken();
  }

  private void parseBlockStatement()
      throws IllegalSyntaxException, IOException {
    {
      boolean isSynchronized = this.skipModifiers();
      if (this.isClassOrInterfaceKeyword()) {
        this.parseClassOrInterfaceDeclaration();
        return;
      }
      if (isSynchronized)
        this.comeTokenBack(TOKEN_SYNCHRONIZED);
    }
    if (JavaTokenUtil.isBasicType(this.currentJavaToken.getKeyword())) {
      this.nextToken();
      this.skipBrackets();
      this.parseVariableDeclarators();
      this.skipSemiColon();
      return;
    }
    if (this.currentJavaToken.getType() != JavaToken.TYPE_IDENTIFIER) {
      this.parseStatement();
      return;
    }
    this.savedJavaToken.copy(this.currentJavaToken);
    this.nextToken();
    boolean isLabel =
        this.currentJavaToken.getOperator() == JavaToken.OPERATOR_COLON;
    this.comeTokenBack(savedJavaToken);
    if (isLabel) {
      this.parseStatement();
      return;
    }
    this.parseQualifiedIdentifier(HANDLING_REFERENCE);
    this.skipBrackets();
    if (this.currentJavaToken.getType() == JavaToken.TYPE_IDENTIFIER) {
      // Local variable declaration
      this.parseVariableDeclarators();
      this.skipSemiColon();
      return;
    }
    this.comeTokenBack(TOKEN_DUMMY_IDENTIFIER);
    this.parseStatement();
  }

  private boolean isClassOrInterfaceKeyword() {
    int keyword = this.currentJavaToken.getKeyword();
    return keyword == JavaToken.KEYWORD_CLASS
        || keyword == JavaToken.KEYWORD_INTERFACE;
  }

  ///////////////////////////////
  // Variable and methods parsing
  ///////////////////////////////

  private void parseVariableDeclarators()
      throws IllegalSyntaxException, IOException {
    while (true) {
      this.skipIdentifier();
      this.skipBrackets();
      if (this.currentJavaToken.getOperator() == JavaToken.OPERATOR_ASSIGNMENT) {
        this.nextToken();
        this.parseVariableInitializer();
      }
      if (this.currentJavaToken.getSeparator() != JavaToken.SEPARATOR_COMMA)
        break;
      this.nextToken();
    }
  }

  private void parseVariableInitializer()
      throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getSeparator()
        == JavaToken.SEPARATOR_START_BRACE_BRACKET)
      this.parseArrayInitializer();
    else
      this.parseExpression();
  }

  private void parseArrayInitializer()
      throws IllegalSyntaxException, IOException {
    this.nextToken();
    if (this.currentJavaToken.getSeparator()
        != JavaToken.SEPARATOR_END_BRACE_BRACKET)
      while (true) {
        this.parseVariableInitializer();
        if (this.currentJavaToken.getSeparator()
            == JavaToken.SEPARATOR_END_BRACE_BRACKET)
          break;
        if (this.currentJavaToken.getSeparator() == JavaToken.SEPARATOR_COMMA) {
          this.nextToken();
          if (this.currentJavaToken.getSeparator()
              == JavaToken.SEPARATOR_END_BRACE_BRACKET)
            break;
          continue;
        }
        this.syntaxErrorCharExpected('}');
      }
    this.nextToken();
  }

  private void parseMemberDeclaration()
      throws IllegalSyntaxException, IOException {
    this.skipModifiers();
    if (this.isClassOrInterfaceKeyword()) {
      this.parseClassDeclaration();
      return;
    }
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_VOID) {
      this.nextToken();
      this.skipIdentifier();
      this.parseMethodDeclaratorRest();
      return;
    }
    if (this.currentJavaToken.getType() == JavaToken.TYPE_IDENTIFIER
        && this.currentJavaToken.valueEquals(
        (String)this.classOrInterfaceIdentifierStack.peek())) {
      // May be a constructor
      savedJavaToken.copy(this.currentJavaToken);
      this.nextToken();
      if (this.currentJavaToken.getSeparator()
          == JavaToken.SEPARATOR_START_ROUND_BRACKET) { // Yes, a constructor
        this.parseMethodDeclaratorRest();
        return;
      }
      this.comeTokenBack(savedJavaToken); // Not a constructor
    }
    this.parseType(HANDLING_REFERENCE);
    this.skipIdentifier();
    if (this.currentJavaToken.getSeparator()
        == JavaToken.SEPARATOR_START_ROUND_BRACKET)
      this.parseMethodDeclaratorRest();
    else {
      this.comeTokenBack(TOKEN_DUMMY_IDENTIFIER);
      this.parseVariableDeclarators();
    }
  }

  private void parseMethodDeclaratorRest()
      throws IllegalSyntaxException, IOException {
    this.parseFormalParameters();
    this.skipBrackets();
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_THROWS) {
      this.nextToken();
      this.parseTypeList();
    }
    if (this.currentJavaToken.getSeparator() == JavaToken.SEPARATOR_SEMI_COLON)
      this.nextToken();
    else
      this.parseBlock();
  }

  private void parseFormalParameters()
      throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getSeparator()
        != JavaToken.SEPARATOR_START_ROUND_BRACKET)
      this.syntaxErrorCharExpected('(');
    this.nextToken();
    if (this.currentJavaToken.getSeparator()
        != JavaToken.SEPARATOR_END_ROUND_BRACKET) {
      this.parseFormalParameter();
      while (this.currentJavaToken.getSeparator() == JavaToken.SEPARATOR_COMMA) {
        this.nextToken();
        this.parseFormalParameter();
      }
      if (this.currentJavaToken.getSeparator()
          != JavaToken.SEPARATOR_END_ROUND_BRACKET)
        this.syntaxErrorCharExpected(')');
    }
    this.nextToken();
  }

  private void parseFormalParameter()
      throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_FINAL)
      this.nextToken();
    this.parseType(HANDLING_REFERENCE);
    this.skipIdentifier();
    this.skipBrackets();
  }

  ////////////////////
  // Statement parsing
  ////////////////////

  private void parseStatement() throws IllegalSyntaxException, IOException {
    this.skipLabels();
    switch (this.currentJavaToken.getSeparator()) {
      case JavaToken.SEPARATOR_SEMI_COLON:
        this.nextToken();
        return;
      case JavaToken.SEPARATOR_START_BRACE_BRACKET:
        this.parseBlock();
        return;
    }
    switch (this.currentJavaToken.getKeyword()) {
      case JavaToken.KEYWORD_IF:
        this.nextToken();
        this.parseParExpression();
        this.parseStatement();
        if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_ELSE) {
          this.nextToken();
          this.parseStatement();
        }
        return;
      case JavaToken.KEYWORD_FOR:
        this.nextToken();
        this.skipSeparator(JavaToken.SEPARATOR_START_ROUND_BRACKET);
        if (this.currentJavaToken.getSeparator()
            != JavaToken.SEPARATOR_SEMI_COLON)
          this.parseForInit();
        this.skipSemiColon();
        if (this.currentJavaToken.getSeparator()
            != JavaToken.SEPARATOR_SEMI_COLON)
          this.parseExpression();
        this.skipSemiColon();
        if (this.currentJavaToken.getSeparator()
            != JavaToken.SEPARATOR_END_ROUND_BRACKET)
          this.parseStatementExpressions();
        this.skipSeparator(JavaToken.SEPARATOR_END_ROUND_BRACKET);
        this.parseStatement();
        return;
      case JavaToken.KEYWORD_WHILE:
        this.nextToken();
        this.parseParExpression();
        this.parseStatement();
        return;
      case JavaToken.KEYWORD_DO:
        this.nextToken();
        this.parseStatement();
        this.skipKeyword(JavaToken.KEYWORD_WHILE);
        this.parseParExpression();
        this.skipSemiColon();
        return;
      case JavaToken.KEYWORD_TRY:
        this.nextToken();
        this.parseBlock();
        while (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_CATCH) {
          this.nextToken();
          this.skipSeparator(JavaToken.SEPARATOR_START_ROUND_BRACKET);
          this.parseFormalParameter();
          this.skipSeparator(JavaToken.SEPARATOR_END_ROUND_BRACKET);
          this.parseBlock();
        }
        if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_FINALLY) {
          this.nextToken();
          this.parseBlock();
        }
        return;
      case JavaToken.KEYWORD_SWITCH:
        this.nextToken();
        this.parseParExpression();
        this.skipSeparator(JavaToken.SEPARATOR_START_BRACE_BRACKET);
        while (true) {
          switch (this.currentJavaToken.getKeyword()) {
            case JavaToken.KEYWORD_CASE:
              this.nextToken();
              this.parseExpression();
              break;
            case JavaToken.KEYWORD_DEFAULT:
              this.nextToken();
              break;
            default:
              this.skipSeparator(JavaToken.SEPARATOR_END_BRACE_BRACKET);
              return;
          }
          this.skipColon();
          while (this.currentJavaToken.getSeparator()
                 != JavaToken.SEPARATOR_END_BRACE_BRACKET
                 && !this.isCaseOrDefault())
            this.parseBlockStatement();
        }
      case JavaToken.KEYWORD_SYNCHRONIZED:
        this.nextToken();
        this.parseParExpression();
        this.parseBlock();
        return;
      case JavaToken.KEYWORD_RETURN:
        this.nextToken();
        if (this.currentJavaToken.getSeparator()
            == JavaToken.SEPARATOR_SEMI_COLON) {
          this.nextToken();
          return;
        }
        this.parseExpression();
        this.skipSemiColon();
        return;
      case JavaToken.KEYWORD_THROW:
        this.nextToken();
        this.parseExpression();
        this.skipSemiColon();
        return;
      case JavaToken.KEYWORD_BREAK:
      case JavaToken.KEYWORD_CONTINUE:
        this.nextToken();
        if (this.currentJavaToken.getType() == JavaToken.TYPE_IDENTIFIER)
          this.nextToken();
        this.skipSemiColon();
        return;
      case JavaToken.KEYWORD_ASSERT:
        this.nextToken();
        this.parseExpression();
        if (this.currentJavaToken.getOperator() == JavaToken.OPERATOR_COLON) {
          this.nextToken();
          this.parseExpression();
        }
        this.skipSemiColon();
        return;
      default:
        this.parseExpression();
        this.skipSemiColon();
    }
  }

  private void skipLabels() throws IllegalSyntaxException, IOException {
    while (true) {
      if (this.currentJavaToken.getType() != JavaToken.TYPE_IDENTIFIER)
        break;
      savedJavaToken.copy(this.currentJavaToken);
      this.nextToken();
      if (this.currentJavaToken.getOperator() != JavaToken.OPERATOR_COLON) {
        this.comeTokenBack(savedJavaToken);
        break;
      }
      this.nextToken();
    }
  }

  private void parseForInit() throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_FINAL)
      this.nextToken();
    if (JavaTokenUtil.isBasicType(this.currentJavaToken.getKeyword())) {
      this.nextToken();
      this.skipBrackets();
      this.parseVariableDeclarators();
      return;
    }
    if (this.currentJavaToken.getType() != JavaToken.TYPE_IDENTIFIER) {
      this.parseStatementExpressions();
      return;
    }
    this.parseQualifiedIdentifier(HANDLING_REFERENCE);
    this.skipBrackets();
    if (this.currentJavaToken.getType() == JavaToken.TYPE_IDENTIFIER) {
      this.parseVariableDeclarators();
      return;
    }
    this.comeTokenBack(TOKEN_DUMMY_IDENTIFIER);
    this.parseStatementExpressions();
  }

  private void parseStatementExpressions()
      throws IllegalSyntaxException, IOException {
    while (true) {
      this.parseExpression();
      if (this.currentJavaToken.getSeparator() != JavaToken.SEPARATOR_COMMA)
        break;
      this.nextToken();
    }
  }

  private boolean isCaseOrDefault() {
    int keyword = this.currentJavaToken.getKeyword();
    return keyword == JavaToken.KEYWORD_CASE
        || keyword == JavaToken.KEYWORD_DEFAULT;
  }

  /////////////////////
  // Expression parsing
  /////////////////////

  private void parseExpression() throws IllegalSyntaxException, IOException {
    this.parseExpression(false);
  }

  private void parseExpression(boolean mayBeType)
      throws IllegalSyntaxException, IOException {
    this.parseExpression1(mayBeType);
    while (JavaTokenUtil.isAssignmentOperator(
        this.currentJavaToken.getOperator())) {
      this.nextToken();
      this.parseExpression1(false);
    }
  }

  private void parseExpression1(boolean mayBeType)
      throws IllegalSyntaxException, IOException {
    this.parseExpression2(mayBeType);
    if (this.currentJavaToken.getOperator() == JavaToken.OPERATOR_QUESTION) {
      this.nextToken();
      this.parseExpression();
      this.skipColon();
      this.parseExpression1(false);
    }
  }

  private void parseExpression2(boolean mayBeType)
      throws IllegalSyntaxException, IOException {
    this.parseExpression2AndHalf(mayBeType);
    while (JavaTokenUtil.isInfixOperator(this.currentJavaToken.getOperator())) {
      this.nextToken();
      this.parseExpression2AndHalf(false);
    }
  }

  private void parseExpression2AndHalf(boolean mayBeType)
      throws IllegalSyntaxException, IOException {
    this.parseExpression3(mayBeType);
    if (this.currentJavaToken.getKeyword() == JavaToken.KEYWORD_INSTANCEOF) {
      this.nextToken();
      this.parseType(HANDLING_REFERENCE);
    }
  }

  private void parseExpression3(boolean mayBeType)
      throws IllegalSyntaxException, IOException {
    boolean isPrimaryRequired = true;
    while (true) {
      if (this.currentJavaToken.getSeparator()
          == JavaToken.SEPARATOR_START_ROUND_BRACKET) {
        this.nextToken();
        if (JavaTokenUtil.isBasicType(this.currentJavaToken.getKeyword())) {
          this.nextToken();
          this.skipBrackets();
          isPrimaryRequired = true;
        }
        else {
          this.parseExpression(true);
          isPrimaryRequired = false;
        }
        this.skipSeparator(JavaToken.SEPARATOR_END_ROUND_BRACKET);
      }
      else if (JavaTokenUtil.isPrefixOperator(
          this.currentJavaToken.getOperator())) {
        this.nextToken();
        isPrimaryRequired = true;
      }
      else
        break;
    }
    if (!this.parsePrimary(mayBeType) & isPrimaryRequired)
      this.syntaxError(ERROR_PRIMARY_EXPECTED);
    this.parseSelectors();
    while (JavaTokenUtil.isPostfixOperator(this.currentJavaToken.getOperator()))
      this.nextToken();
  }

  private boolean parsePrimary(boolean mayBeType)
      throws IllegalSyntaxException, IOException {
    switch (this.currentJavaToken.getType()) {
      case JavaToken.TYPE_IDENTIFIER:
        this.parseQualifiedIdentifier(HANDLING_REFERENCE);
        this.parseIdentifierSuffix(mayBeType);
        break;
      case JavaToken.TYPE_KEYWORD: {
        int keyword = this.currentJavaToken.getKeyword();
        this.savedJavaToken.copy(this.currentJavaToken);
        this.nextToken();
        if (JavaTokenUtil.isBasicType(keyword)) {
          this.skipBrackets();
          this.skipSeparator(JavaToken.SEPARATOR_DOT);
          this.skipKeyword(JavaToken.KEYWORD_CLASS);
          break;
        }
        switch (keyword) {
          case JavaToken.KEYWORD_THIS:
            if (this.currentJavaToken.getSeparator()
                == JavaToken.SEPARATOR_START_ROUND_BRACKET)
              this.parseArguments();
            break;
          case JavaToken.KEYWORD_SUPER:
            this.parseSuperSuffix();
            break;
          case JavaToken.KEYWORD_NEW:
            this.parseTypeWithoutBrackets(HANDLING_INSTANCIATION);
            if (this.currentJavaToken.getSeparator()
                == JavaToken.SEPARATOR_START_SQUARE_BRACKET)
              this.parseArrayCreatorRest();
            else
              this.parseClassCreatorRest();
            break;
          case JavaToken.KEYWORD_VOID:
            this.skipSeparator(JavaToken.SEPARATOR_DOT);
            this.skipKeyword(JavaToken.KEYWORD_CLASS);
            break;
          default:
            this.comeTokenBack(savedJavaToken);
            return false;
        }
        break;
      }
      case JavaToken.TYPE_LITERAL:
        this.nextToken();
        break;
      case JavaToken.TYPE_SEPARATOR:
        if (this.currentJavaToken.getSeparator()
            == JavaToken.SEPARATOR_START_ROUND_BRACKET) {
          this.parseParExpression();
          break;
        }
      default:
        return false;
    }
    return true;
  }

  private void parseIdentifierSuffix(boolean mayBeTypeSuffix)
      throws IllegalSyntaxException, IOException {
    switch (this.currentJavaToken.getSeparator()) {
      case JavaToken.SEPARATOR_START_SQUARE_BRACKET:
        this.nextToken();
        if (this.currentJavaToken.getSeparator()
            == JavaToken.SEPARATOR_END_SQUARE_BRACKET) {
          this.nextToken();
          this.skipBrackets();
          if (!mayBeTypeSuffix ||
              this.currentJavaToken.getSeparator() == JavaToken.SEPARATOR_DOT) {
            this.skipSeparator(JavaToken.SEPARATOR_DOT);
            this.skipKeyword(JavaToken.KEYWORD_CLASS);
          }
        }
        else {
          this.parseExpression();
          this.skipSeparator(JavaToken.SEPARATOR_END_SQUARE_BRACKET);
        }
        break;
      case JavaToken.SEPARATOR_START_ROUND_BRACKET:
        this.parseArguments();
        break;
      case JavaToken.SEPARATOR_DOT:
        this.nextToken();
        switch (this.currentJavaToken.getKeyword()) {
          case JavaToken.KEYWORD_CLASS:
          case JavaToken.KEYWORD_THIS:
            this.nextToken();
            break;
          case JavaToken.KEYWORD_SUPER:
            this.nextToken();
            this.parseArguments();
            break;
          case JavaToken.KEYWORD_NEW:
            this.nextToken();
            this.parseInnerCreator();
            break;
          default:
            this.syntaxError(ERROR_IDENTIFIER_SUFFIX_EXPECTED);
        }
    }
  }

  private void parseSelectors() throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getSeparator()
        == JavaToken.SEPARATOR_START_ROUND_BRACKET)
      this.parseArguments();
    while (true) {
      switch (this.currentJavaToken.getSeparator()) {
        case JavaToken.SEPARATOR_DOT:
          this.nextToken();
          switch (this.currentJavaToken.getKeyword()) {
            case JavaToken.KEYWORD_THIS:
              this.nextToken();
              break;
            case JavaToken.KEYWORD_SUPER:
              this.nextToken();
              this.parseSuperSuffix();
              break;
            case JavaToken.KEYWORD_NEW:
              this.nextToken();
              this.parseInnerCreator();
              break;
            default:
              this.skipIdentifier();
              if (this.currentJavaToken.getSeparator()
                  == JavaToken.SEPARATOR_START_ROUND_BRACKET)
                this.parseArguments();
          }
          break;
        case JavaToken.SEPARATOR_START_SQUARE_BRACKET:
          this.nextToken();
          this.parseExpression();
          this.skipSeparator(JavaToken.SEPARATOR_END_SQUARE_BRACKET);
          break;
        default:
          return;
      }
    }
  }

  private void parseSuperSuffix() throws IllegalSyntaxException, IOException {
    switch (this.currentJavaToken.getSeparator()) {
      case JavaToken.SEPARATOR_START_ROUND_BRACKET:
        this.parseArguments();
        break;
      case JavaToken.SEPARATOR_DOT:
        this.nextToken();
        this.skipIdentifier();
        if (this.currentJavaToken.getSeparator()
            == JavaToken.SEPARATOR_START_ROUND_BRACKET)
          this.parseArguments();
        break;
      default:
        this.syntaxError(ERROR_SUPER_SUFFIX_EXPECTED);
    }
  }

  private void parseArrayCreatorRest()
      throws IllegalSyntaxException, IOException {
    this.skipSeparator(JavaToken.SEPARATOR_START_SQUARE_BRACKET);
    if (this.currentJavaToken.getSeparator()
        == JavaToken.SEPARATOR_END_SQUARE_BRACKET) {
      this.nextToken();
      this.skipBrackets();
      this.parseArrayInitializer();
    }
    else {
      this.parseExpression();
      this.skipSeparator(JavaToken.SEPARATOR_END_SQUARE_BRACKET);
      boolean isDimensioned = true;
      while (this.currentJavaToken.getSeparator()
             == JavaToken.SEPARATOR_START_SQUARE_BRACKET) {
        this.nextToken();
        if (this.currentJavaToken.getSeparator()
            == JavaToken.SEPARATOR_END_SQUARE_BRACKET)
          isDimensioned = false;
        else if (isDimensioned)
          this.parseExpression();
        else
          this.syntaxErrorCharExpected(']');
        this.skipSeparator(JavaToken.SEPARATOR_END_SQUARE_BRACKET);
      }
    }
  }

  private void parseClassCreatorRest()
      throws IllegalSyntaxException, IOException {
    this.parseArguments();
    if (this.currentJavaToken.getSeparator()
        == JavaToken.SEPARATOR_START_BRACE_BRACKET)
      this.parseClassOrInterfaceBody();
  }

  private void parseInnerCreator() throws IllegalSyntaxException, IOException {
    this.skipIdentifier();
    this.parseClassCreatorRest();
  }

  private void parseArguments() throws IllegalSyntaxException, IOException {
    this.skipSeparator(JavaToken.SEPARATOR_START_ROUND_BRACKET);
    if (this.currentJavaToken.getSeparator()
        != JavaToken.SEPARATOR_END_ROUND_BRACKET) {
      this.parseExpression();
      while (this.currentJavaToken.getSeparator() == JavaToken.SEPARATOR_COMMA)
      {
        this.nextToken();
        this.parseExpression();
      }
    }
    this.skipSeparator(JavaToken.SEPARATOR_END_ROUND_BRACKET);
  }

  private void parseParExpression() throws IllegalSyntaxException, IOException {
    this.skipSeparator(JavaToken.SEPARATOR_START_ROUND_BRACKET);
    this.parseExpression();
    this.skipSeparator(JavaToken.SEPARATOR_END_ROUND_BRACKET);
  }

  ////////////////////////
  // Type relative methods
  ////////////////////////

  /**
   * Return the full qualified identifier of a outer class or interface.
   *   whose reference migth be given by a qualified identifier
   *   with this format :
   *
   * qualifiedIdentifier ::= (( packageIdentifier '.' )* outerClassIdentifier
   *   ( '.' innerClassIdentifier )* ( '.' fieldIdentifier )*
   *   ( '.' methodIdentifier )?)
   *   | ( fieldIdentifier ( '.' fieldIdentifier )* ( '.' methodIdentifier )?)
   *
   * Return null if the outer class or interface is absent or unknown.
   *
   * It is supposed that :
   * - all class identifiers begin with an upper case,
   * - no package identifier begins with an upper case,
   * - no method identifier begins with an upper case,
   * - no variable field identifier begins with an upper case.
   */
  private String fullClassOrInterfaceQualifiedIdentifier(
      String qualifiedIdentifier) {
    int dotPos = qualifiedIdentifier.indexOf('.');
    String firstIdentifier = dotPos >= 0 ?
        qualifiedIdentifier.substring(0, dotPos) : qualifiedIdentifier;
    if (Character.isUpperCase(qualifiedIdentifier.charAt(0))) {
      // The first identifier is supposed to reference an outer class
      return (String)this.
          fromClassOrInterfaceIdentifierToClassOrInterfaceQualifiedIdentifier.
          get(firstIdentifier);
    }
    // The first identifier is supposed to reference a primary package
    if (this.primaryPackageIdentifierSet != null
        && !this.primaryPackageIdentifierSet.contains(firstIdentifier))
      return null;
    int index = 0;
    int length = qualifiedIdentifier.length();
    while (true) {
      dotPos = qualifiedIdentifier.indexOf('.', index);
      if (Character.isUpperCase(qualifiedIdentifier.charAt(index))) {
        if (dotPos >= 0)
          length = dotPos;
        break;
      }
      if (dotPos < 0)
        return null;
      index = dotPos + 1;
    }
    return length == qualifiedIdentifier.length() ? qualifiedIdentifier :
        qualifiedIdentifier.substring(0, length);
  }

  private static final int HANDLING_NO = 0;
  private static final int HANDLING_REFERENCE = 1;
  private static final int HANDLING_INSTANCIATION = 2;

  /**
   * Parse a type and return the qualified identifier of the used class
   *   or interface if any, null elsewhere.
   * handling = HANDLING_*
   */
  private String parseType(int handling)
      throws IllegalSyntaxException, IOException {
    String qualifiedIdentifier =
        this.parseTypeWithoutBrackets(handling);
    this.skipBrackets();
    return qualifiedIdentifier;
  }

  /**
   * Parse a type without brackets and return the qualified identifier
   *   of the used class or interface if any, null elsewhere.
   * handling = HANDLING_*
   */
  private String parseTypeWithoutBrackets(int handling)
      throws IllegalSyntaxException, IOException {
    if (JavaTokenUtil.isBasicType(this.currentJavaToken.getKeyword())) {
      this.nextToken();
      return null;
    }
    return this.parseQualifiedIdentifier(handling);
  }

  /**
   * Parse a list of types and return the list of qualified identifiers of the
   *   used classes and interfaces.
   * handling = HANDLING_*
   */
  private List parseTypeList(int handling)
      throws IllegalSyntaxException, IOException {
    String qualifiedIdentifier = this.parseType(handling);
    if (this.currentJavaToken.getSeparator() != JavaToken.SEPARATOR_COMMA)
      return qualifiedIdentifier != null ?
          Collections.singletonList(qualifiedIdentifier) :
          Collections.EMPTY_LIST;
    this.nextToken();
    List typeList = new ArrayList();
    if (qualifiedIdentifier != null)
      typeList.add(qualifiedIdentifier);
    while (true) {
      qualifiedIdentifier = this.parseType(handling);
      if (qualifiedIdentifier != null)
        typeList.add(qualifiedIdentifier);
      if (this.currentJavaToken.getSeparator() != JavaToken.SEPARATOR_COMMA)
        break;
      this.nextToken();
    }
    return typeList;
  }

  /**
   * Parse a list of types.
   */
  private void parseTypeList() throws IllegalSyntaxException, IOException {
    while (true) {
      this.parseType(HANDLING_REFERENCE);
      if (this.currentJavaToken.getSeparator() != JavaToken.SEPARATOR_COMMA)
        break;
      this.nextToken();
    }
  }

  /**
   * Parses and returns a qualified identifier and handles the referenced class
   *   or interface if any.
   * handling = HANDLING_*
   */
  private String parseQualifiedIdentifier(int handling)
      throws IllegalSyntaxException, IOException {
    String qualifiedIdentifier = this.parseQualifiedIdentifier();
    qualifiedIdentifier =
        this.fullClassOrInterfaceQualifiedIdentifier(qualifiedIdentifier);
    if (qualifiedIdentifier != null)
      switch (handling) {
        case HANDLING_REFERENCE:
          this.compilationUnitHandler.
              classOrInterfaceReference(qualifiedIdentifier);
          break;
        case HANDLING_INSTANCIATION:
          this.compilationUnitHandler.instanciation(qualifiedIdentifier);
          break;
      }
    return qualifiedIdentifier;
  }

  //////////////////////////
  // General parsing methods
  //////////////////////////

  /**
   * Parse and return an identifier.
   */
  private String parseIdentifier() throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getType() != JavaToken.TYPE_IDENTIFIER)
      this.syntaxError(ERROR_IDENTIFIER_EXPECTED);
    String identifier = this.currentJavaToken.getValue();
    this.nextToken();
    return identifier;
  }

  /**
   * Parse and return a qualified identifier.
   */
  private String parseQualifiedIdentifier()
      throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getType() != JavaToken.TYPE_IDENTIFIER)
      this.syntaxError(ERROR_IDENTIFIER_EXPECTED);
    StringBuffer sb = new StringBuffer();
    while (true) {
      this.currentJavaToken.appendValueTo(sb);
      this.nextToken();
      if (this.currentJavaToken.getSeparator() != JavaToken.SEPARATOR_DOT)
        break;
      this.nextToken();
      if (this.currentJavaToken.getType() != JavaToken.TYPE_IDENTIFIER) {
        this.comeTokenBack(TOKEN_DOT);
        break;
      }
      sb.append('.');
    }
    return sb.toString();
  }

  private void skipIdentifier() throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getType() != JavaToken.TYPE_IDENTIFIER)
      this.syntaxError(ERROR_IDENTIFIER_EXPECTED);
    this.nextToken();
  }

  private void skipSeparator(int separator)
      throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getSeparator() == separator)
      this.nextToken();
    else
      switch (separator) {
        case JavaToken.SEPARATOR_START_ROUND_BRACKET:
          this.syntaxErrorCharExpected('(');
        case JavaToken.SEPARATOR_END_ROUND_BRACKET:
          this.syntaxErrorCharExpected(')');
        case JavaToken.SEPARATOR_START_BRACE_BRACKET:
          this.syntaxErrorCharExpected('{');
        case JavaToken.SEPARATOR_END_BRACE_BRACKET:
          this.syntaxErrorCharExpected('}');
        case JavaToken.SEPARATOR_START_SQUARE_BRACKET:
          this.syntaxErrorCharExpected('[');
        case JavaToken.SEPARATOR_END_SQUARE_BRACKET:
          this.syntaxErrorCharExpected(']');
        case JavaToken.SEPARATOR_SEMI_COLON:
          this.syntaxErrorCharExpected(';');
        case JavaToken.SEPARATOR_COMMA:
          this.syntaxErrorCharExpected(',');
        case JavaToken.SEPARATOR_DOT:
          this.syntaxErrorCharExpected('.');
        default:
          throw new IllegalArgumentException();
      }
  }

  private void skipSemiColon() throws IllegalSyntaxException, IOException {
    this.skipSeparator(JavaToken.SEPARATOR_SEMI_COLON);
  }

  private void skipColon() throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getOperator() != JavaToken.OPERATOR_COLON)
      this.syntaxErrorCharExpected(':');
    this.nextToken();
  }

  private void skipKeyword(int keyword)
      throws IllegalSyntaxException, IOException {
    if (this.currentJavaToken.getKeyword() == keyword)
      this.nextToken();
    else
      switch (keyword) {
        case JavaToken.KEYWORD_CASE:
          this.syntaxErrorKeywordExpected("case");
        case JavaToken.KEYWORD_CATCH:
          this.syntaxErrorKeywordExpected("catch");
        case JavaToken.KEYWORD_DEFAULT:
          this.syntaxErrorKeywordExpected("default");
        case JavaToken.KEYWORD_DO:
          this.syntaxErrorKeywordExpected("do");
        case JavaToken.KEYWORD_ELSE:
          this.syntaxErrorKeywordExpected("else");
        case JavaToken.KEYWORD_WHILE:
          this.syntaxErrorKeywordExpected("while");
        default:
          throw new IllegalArgumentException();
      }
  }

  /**
   * Skip zero or more modifiers.
   * Return true is the last one is "synchronized".
   */
  private boolean skipModifiers() throws IllegalSyntaxException, IOException {
    boolean isLastOneSynchronized = false;
    while (true) {
      int keyword = this.currentJavaToken.getKeyword();
      if (!JavaTokenUtil.isModifier(keyword))
        break;
      isLastOneSynchronized = keyword == JavaToken.KEYWORD_SYNCHRONIZED;
      this.nextToken();
    }
    return isLastOneSynchronized;
  }

  private void skipBrackets() throws IllegalSyntaxException, IOException {
    while (this.currentJavaToken.getSeparator()
           == JavaToken.SEPARATOR_START_SQUARE_BRACKET) {
      this.nextToken();
      if (this.currentJavaToken.getSeparator()
           == JavaToken.SEPARATOR_END_SQUARE_BRACKET)
         this.nextToken();
       else {
         this.comeTokenBack(TOKEN_START_SQUARE_BRACKET);
         break;
       }
    }
  }

  ////////////////
  // Token methods
  ////////////////

  private void nextToken() throws IllegalSyntaxException, IOException {
    if (this.javaTokenStack.isEmpty())
      this.currentJavaToken = this.javaTokenizer.nextToken();
    else
      this.currentJavaToken = (JavaToken)this.javaTokenStack.pop();
  }

  private void comeTokenBack(JavaToken javaToken) {
    if (javaToken == null)
      throw new IllegalArgumentException();
    if (this.currentJavaToken != null)
      this.javaTokenStack.push(this.currentJavaToken);
    this.currentJavaToken = javaToken;
  }

  ///////////
  // Warnings
  ///////////

  private static final String WARNING_PACKAGE_NAME =
      "A package identifier should not begin with an upper case";
  private static final String WARNING_CLASS_OR_INTERFACE_NAME =
      "A class or interface identifier should begin with an upper case";
  private static final String WARNING_UNKNOWN_PACKAGE = "Unknown package";

  private void warning(String message) {
    if (this.isWarningPrinted) {
      Localizer localizer = this.javaTokenizer.getLocalizer();
      System.out.println("Warning in " + localizer.getPathName()
                         + " at line " + localizer.getLine()
                         + ", row " + localizer.getRow() + " :");
      System.out.println(message);
    }
  }

  /**
   * A class or interface identifier must begin with an upper case.
   */
  private void verifyCurrentClassOrInterfaceIdentifier() {
    if (!Character.isUpperCase(
        ((String)this.classOrInterfaceIdentifierStack.peek()).charAt(0)))
      warning(WARNING_CLASS_OR_INTERFACE_NAME);
  }

  /**
   * A package identifier must not begin with an upper case.
   */
  private void verifyPackageQualifiedIdentifier(String qualifiedIdentifier) {
    int index = 0;
    while (true) {
      if (Character.isUpperCase(qualifiedIdentifier.charAt(index))) {
        warning(WARNING_PACKAGE_NAME);
        return;
      }
      int dotPos = qualifiedIdentifier.indexOf('.', index);
      if (dotPos < 0)
        return;
      index = dotPos + 1;
    }
  }

  /**
   * A package identifier must not begin with an upper case.
   * A class or interface identifier must begin with an upper case.
   */
  private void verifyImportQualifiedIdentifier(String qualifiedIdentifier) {
    int index = 0;
    while (true) {
      boolean isClassOrInterfaceIdentifier =
          Character.isUpperCase(qualifiedIdentifier.charAt(index));
      int dotPos = qualifiedIdentifier.indexOf('.', index);
      if (dotPos < 0) {
        if (!isClassOrInterfaceIdentifier)
          warning(WARNING_CLASS_OR_INTERFACE_NAME);
        return;
      }
      if (isClassOrInterfaceIdentifier) {
        warning(WARNING_PACKAGE_NAME);
        return;
      }
      index = dotPos + 1;
    }
  }

  /////////
  // Errors
  /////////

  private static final String ERROR_IDENTIFIER_EXPECTED =
      "Identifier expected";
  private static final String ERROR_CLASS_OR_INTERFACE_EXPECTED =
      "\"class\" or \"interface\" expected";
  private static final String ERROR_ILLEGAL_LABEL = "Illegal label";
  private static final String ERROR_PRIMARY_EXPECTED =
      "Primary expected";
  private static final String ERROR_IDENTIFIER_SUFFIX_EXPECTED =
      "Identifier suffix expected";
  private static final String ERROR_SUPER_SUFFIX_EXPECTED =
      "Super suffix expected";

  private void syntaxError(String message) throws IllegalSyntaxException {
    Localizer localizer = this.javaTokenizer.getLocalizer();
    throw new IllegalSyntaxException(
        localizer.getPathName(), localizer.getLine(), localizer.getRow(),
        message);
  }

  private void syntaxErrorCharExpected(char expectedChar)
      throws IllegalSyntaxException {
    this.syntaxError("\"" + expectedChar + "\" expected");
  }

  private void syntaxErrorKeywordExpected(String expectedKeyword)
      throws IllegalSyntaxException {
    this.syntaxError("\"" + expectedKeyword + "\" expected");
  }

  private void trace(int k) {
    System.out.println("$" + k + ":" + this.currentJavaToken.getValue());
  }

  /////////////////
  // Static methods
  /////////////////

  private static final String getLastIdentifier(String qualifiedIdentifier) {
    int dotPos = qualifiedIdentifier.lastIndexOf('.');
    return dotPos < 0 ? qualifiedIdentifier :
        qualifiedIdentifier.substring(dotPos + 1);
  }
}
