/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
/// Java is a trademark of Sun Microsystems, Inc.
package org.xmloperator.java.syntax;

import java.util.Set;
import java.util.HashSet;
import java.util.Arrays;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.Reader;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import org.xmloperator.java.syntax.IllegalSyntaxException;
import org.xmloperator.java.syntax.Localizer;
import org.xmloperator.java.syntax.JavaTokenizer;

public class JavaTokenizerImpl implements JavaTokenizer, Localizer {

  private static final String ERROR_UNCLOSED_COMMENT = "Unclosed comment";
  private static final String ERROR_ILLEGAL_CHARACTER = "Illegal character";
  private static final String ERROR_TOO_LONG_IDENTIFIER = "Too long identifier";
  private static final String ERROR_UNCLOSED_CHARACTER_LITERAL =
      "Unclosed character literal";
  private static final String ERROR_ILLEGAL_ESCAPE_CHARACTER =
      "Illegal escape character";
  private static final String[] keywordsAndNamedLiterals =
      {"false", "true", "null",
      "abstract", "boolean", "break", "byte", "case", "catch", "char", "class",
      "const", "continue", "default", "do", "double", "else", "extends",
      "final", "finally", "float", "for", "goto", "if", "implements", "import",
      "instanceof", "int", "interface", "long", "native", "new", "package",
      "private", "protected", "public", "return", "short", "static", "strictfp",
      "super", "switch", "synchronized", "this", "throw", "throws", "transient",
      "try", "void", "volatile", "while", "assert"};
  private static final int keyWordCodingBitCount = 10;
  private static final int[] signatureToKeywordOrNamedLiteralIndex =
      new int[1 << keyWordCodingBitCount];
  private static final char[] separatorsAndNumericStarts =
      {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
      '(', ')', '{', '}', '[', ']', ';', ',', '.'};
  private static final int[] signatureToSeparatorOrNumericStartIndex =
      new int[0x80];
  private static final char[] hexadecimalChars =
      {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
      'a', 'b', 'c', 'd', 'e', 'f', 'A', 'B', 'C', 'D', 'E', 'F'};
  private static final int[] signatureToHexadecimalCharIndex =
      new int[0x80];

  private PrefetchReader prefetchReader;
  private final JavaTokenImpl currentToken = new JavaTokenImpl();
  private final Buffer buffer = new Buffer();

  private static final class Buffer {
    public final char[] buffer = new char[4096];
    public int length = 0;

    public boolean add(char ch) {
      if (this.length >= this.buffer.length)
        return false;
      this.buffer[this.length++] = ch;
      return true;
    }
  }

  static {
    Arrays.fill(signatureToKeywordOrNamedLiteralIndex, -1);
    Set signatureSet = new HashSet();
    for (int k = 0; k < keywordsAndNamedLiterals.length; k++) {
      String value = keywordsAndNamedLiterals[k];
      int signature = keywordSignature(value);
      //System.err.println(value + ":" + signature);
      if (!signatureSet.add(new Integer(signature))) {
        System.err.println("Bad keywords differencing");
        System.exit(1);
      }
      signatureToKeywordOrNamedLiteralIndex[signature] = k;
    }
    Arrays.fill(signatureToSeparatorOrNumericStartIndex, -1);
    for (int k = 0; k < separatorsAndNumericStarts.length; k++)
      signatureToSeparatorOrNumericStartIndex
          [(int)separatorsAndNumericStarts[k] & 0x7F] = k;
    Arrays.fill(signatureToHexadecimalCharIndex, -1);
    for (int k = 0; k < hexadecimalChars.length; k++)
      signatureToHexadecimalCharIndex
          [(int)hexadecimalChars[k] & 0x7F] = k;
  }

  private static final int keywordSignature(String value) {
    int signature = 0;
    for (int k = 0; k < value.length(); k++)
      signature += signature + signature + (int)value.charAt(k) + k;
    return signature & ((1 << keyWordCodingBitCount) - 1);
  }

  private static final int keywordSignature(char[] chars, int start, int length)
  {
    int signature = 0;
    for (int k = 0; k < length; k++)
      signature += signature + signature + (int)chars[start + k] + k;
    return signature & ((1 << keyWordCodingBitCount) - 1);
  }

  private static char[] equalityBuffer = new char[16];

  private static final boolean equals(String value,
                                      char[] chars, int start, int length) {
    if (value.length() != length)
      return false;
    if (length > equalityBuffer.length)
      throw new IllegalArgumentException();
    value.getChars(0, length, equalityBuffer, 0);
    for (int k = 0; k < length; k++)
      if (chars[start + k] != equalityBuffer[k])
        return false;
    return true;
  }

  private static class PrefetchReader {
    private Reader reader;
    private String pathName;
    private int line = 1;
    private int row = 0;
    private int prefetch1 = -1;
    private int prefetch2 = -1;

    public PrefetchReader(Reader reader, String pathName) {
      this.reader = reader;
      this.pathName = pathName;
    }

    public int get() throws IOException, IllegalSyntaxException {
      if (this.prefetch1 < 0) {
        this.prefetch1 = this.next();
        if (this.prefetch1 == '\\') {
          this.prefetch2 = this.next();
          if (this.prefetch2 == 'u') {
            // Unicode escape
            this.prefetch1 = 0;
            for (int k = 0; k < 4; k++) {
              int hexadecimalCharIndex =
                  signatureToHexadecimalCharIndex[this.next() & 0x7F];
              if (hexadecimalCharIndex < 0)
                throw new IllegalSyntaxException(
                    this.pathName, this.line, this.row,
                    ERROR_ILLEGAL_ESCAPE_CHARACTER);
              this.prefetch1 <<= 4;
              this.prefetch1 += hexadecimalCharIndex;
            }
            this.prefetch2 = -1;
          }
        }
      }
      return this.prefetch1;
    }

    public int read() throws IOException, IllegalSyntaxException {
      int ch = this.get();
      this.prefetch1 = this.prefetch2;
      this.prefetch2 = -1;
      return ch;
    }

    private int next() throws IOException {
      int ch = this.reader.read();
      if (ch >= 0) {
        this.row++;
        if (ch == '\n') {
          this.line++;
          this.row = 0;
        }
      }
      return ch;
    }

    public String getPathName() {
      return this.pathName;
    }

    public int getLine() {
      return this.line;
    }

    public int getRow() {
      return this.row;
    }

    public void close() throws IOException {
      this.reader.close();
      this.prefetch1 = -1;
    }
  }

  public JavaTokenizerImpl(String pathName, InputStream inputStream) {
    Reader reader = new BufferedReader(new InputStreamReader(inputStream));
    this.prefetchReader = new PrefetchReader(reader, pathName);
  }

  private void syntaxError(String message) throws IllegalSyntaxException {
    throw new IllegalSyntaxException(
        this.getPathName(), this.getLine(), this.getRow(), message);
  }

  private int readCharacterLiteral()
      throws IllegalSyntaxException, IOException {
    int ch = this.prefetchReader.read();
    if (ch < 0 || ch == '\r' || ch == '\n')
      this.syntaxError(ERROR_UNCLOSED_CHARACTER_LITERAL);
    this.buffer.add((char)ch);
    return ch;
  }

  private int readChar() throws IOException, IllegalSyntaxException {
    int ch = this.prefetchReader.read();
    this.buffer.add((char)ch);
    return ch;
  }

  private void readDigits() throws IOException, IllegalSyntaxException {
    while (true) {
      int ch = this.prefetchReader.get();
      if (ch < '0' || ch > '9')
        break;
      this.readChar();
    }
  }

  private static final boolean isIntegerTypeSuffix(char ch) {
    return ch == 'L' || ch == 'l';
  }

  private static final boolean isFloatingPointTypeSuffix(char ch) {
    return ch == 'f' || ch == 'F' || ch == 'd' || ch == 'D';
  }

  ///////////////////////////
  // Localizer implementation
  ///////////////////////////

  public String getPathName() {
    return this.prefetchReader.getPathName();
  }

  public int getLine() {
    return this.prefetchReader.getLine();
  }

  public int getRow() {
    return this.prefetchReader.getRow();
  }

  ///////////////////////////////
  // JavaTokenizer implementation
  ///////////////////////////////

  public JavaToken nextToken() throws IllegalSyntaxException, IOException {
    this.currentToken.clear();
    int ch;
    while (true) {
      ch = this.prefetchReader.read();
      if (ch < 0)
        return this.currentToken;
      if (Character.isWhitespace((char)ch))
        continue;
      if (ch != '/')
        break;
      int ch1 = this.prefetchReader.get();
      if (ch1 < 0)
        break;
      if (ch1 == '/') {
        // EndOfLineComment
        this.prefetchReader.read();
        while (true) {
          ch = this.prefetchReader.read();
          if (ch < 0)
            return this.currentToken;
          if (ch == '\n' || ch == '\r')
            break;
        }
        continue;
      }
      else if (ch1 == '*') {
        // TraditionalComment
        this.prefetchReader.read();
        while (true) {
          ch = this.prefetchReader.read();
          if (ch < 0)
            this.syntaxError(ERROR_UNCLOSED_COMMENT);
          if (ch != '*')
            continue;
          ch1 = this.prefetchReader.get();
          if (ch1 < 0)
            this.syntaxError(ERROR_UNCLOSED_COMMENT);
          if (ch1 == '/')
            break;
        }
        this.prefetchReader.read();
        continue;
      }
      else
        break;
    }
    // Token start
    if (Character.isJavaIdentifierStart((char)ch)) {
      // Identifier or keyword or boolean literal or null literal
      this.buffer.length = 0;
      this.buffer.add((char)ch);
      while (true) {
        {
          int ch1 = this.prefetchReader.get();
          if (ch1 < 0 || !Character.isJavaIdentifierPart( (char) ch1))
            break;
        }
        ch = this.prefetchReader.read();
        if (!this.buffer.add((char)ch))
          this.syntaxError(ERROR_TOO_LONG_IDENTIFIER);
      }
      int keywordOrNamedLiteralIndex = signatureToKeywordOrNamedLiteralIndex[
          keywordSignature(this.buffer.buffer, 0, this.buffer.length)];
      if (keywordOrNamedLiteralIndex < 0
          || !equals(keywordsAndNamedLiterals[keywordOrNamedLiteralIndex],
                     this.buffer.buffer, 0, this.buffer.length))
        this.currentToken.setIdentifier(this.buffer.buffer, 0,
                                        this.buffer.length);
      else if (keywordOrNamedLiteralIndex >= 3)
        this.currentToken.setKeyword(keywordOrNamedLiteralIndex - 3);
      else
        this.currentToken.setLiteral(this.buffer.buffer, 0, this.buffer.length);
      return this.currentToken;
    }
    // Literal or separator or operator
    int separatorOrNumericStartIndex =
        signatureToSeparatorOrNumericStartIndex[ch & 0x7F];
    if (separatorOrNumericStartIndex >= 0
        && ch == separatorsAndNumericStarts[separatorOrNumericStartIndex]) {
      // Separator or numeric
      boolean isNumeric;
      if (ch == '.') {
        // Floating point or separator ?
        int ch1 = this.prefetchReader.get();
        isNumeric = ch1 >= '0' && ch1 <= '9';
      }
      else
        isNumeric = separatorOrNumericStartIndex < 10;
      if (isNumeric) {
        // Integer or floating point
        this.buffer.length = 0;
        this.buffer.add((char)ch);
        boolean isHexadecimal;
        if (ch == '0') {
          int ch1 = this.prefetchReader.get();
          isHexadecimal = ch1 == 'x' || ch1 == 'X';
        }
        else
          isHexadecimal = false;
        if (isHexadecimal) {
          ch = this.prefetchReader.read();
          this.buffer.add((char)ch);
          while (true) {
            int ch1 = this.prefetchReader.get();
            if (ch1 < 0)
             break;
            int hexadecimalCharIndex =
                signatureToHexadecimalCharIndex[ch1 & 0x7F];
            if (hexadecimalCharIndex < 0
                || ch1 != hexadecimalChars[hexadecimalCharIndex])
              break;
            ch = this.prefetchReader.read();
            this.buffer.add((char)ch);
          }
          if (isIntegerTypeSuffix( (char)this.prefetchReader.get()))
            this.readChar();
        }
        else {
          // Decimal, octal or floating point
          if (ch != '.') {
            this.readDigits();
            if (this.prefetchReader.get() == '.')
              ch = this.readChar();
          }
          if (ch == '.')
            this.readDigits();
          int ch1 = this.prefetchReader.get();
          if (ch1 == 'e' || ch1 == 'E') {
            // Exponant part
            this.readChar();
            ch1 = this.prefetchReader.get();
            if (ch1 == '+' || ch1 == '-')
              this.readChar();
            this.readDigits();
          }
          ch1 = this.prefetchReader.get();
          if (isIntegerTypeSuffix((char)ch1)
              || isFloatingPointTypeSuffix((char)ch1))
            this.readChar();
        }
        this.currentToken.setLiteral(this.buffer.buffer, 0, this.buffer.length);
      }
      else
        this.currentToken.setSeparator(separatorOrNumericStartIndex - 10);
      return this.currentToken;
    }
    if (ch == '"') {
      // String literal
      this.buffer.length = 0;
      this.buffer.add((char)ch);
      while (true) {
        ch = this.readCharacterLiteral();
        if (ch == '\\') {
          ch = this.readCharacterLiteral();
          continue;
        }
        if (ch == '"')
          break;
      }
      this.currentToken.setLiteral(this.buffer.buffer, 0, this.buffer.length);
      return this.currentToken;
    }
    if (ch == '\'') {
      // Character literal
      this.buffer.length = 0;
      this.buffer.add((char)ch);
      ch = this.readCharacterLiteral();
      if (ch == '\\') {
        ch = this.readCharacterLiteral();
        if (ch >= '0' && ch <= '7')
          do {
            ch = this.readCharacterLiteral();
          } while (ch >= '0' && ch <= '7');
        else
          ch = this.readCharacterLiteral();
      }
      else
        ch = this.readCharacterLiteral();
      if (ch != '\'')
        this.syntaxError(ERROR_UNCLOSED_CHARACTER_LITERAL);
      this.currentToken.setLiteral(this.buffer.buffer, 0, this.buffer.length);
      return this.currentToken;
    }
    // Operator
    int ch1 = this.prefetchReader.get();
    int operator;
    switch (ch) {
      case '=':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_EQUAL;
        }
        else
          operator = JavaToken.OPERATOR_ASSIGNMENT;
        break;
      case '!':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_NOT_EQUAL;
        }
        else
          operator = JavaToken.OPERATOR_NOT;
        break;
      case '<':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_LESSER_OR_EQUAL;
        }
        else if (ch1 == '<') {
          this.prefetchReader.read();
          ch1 = this.prefetchReader.get();
          if (ch1 == '=') {
            this.prefetchReader.read();
            operator = JavaToken.OPERATOR_LEFT_SHIFT_ASSIGNMENT;
          }
          else
            operator = JavaToken.OPERATOR_LEFT_SHIFT;
        }
        else
          operator = JavaToken.OPERATOR_LESSER;
        break;
      case '>':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_GREATER_OR_EQUAL;
        }
        else if (ch1 == '>') {
          this.prefetchReader.read();
          ch1 = this.prefetchReader.get();
          if (ch1 == '=') {
            this.prefetchReader.read();
            operator = JavaToken.OPERATOR_RIGHT_SHIFT_ASSIGNMENT;
          }
          else if (ch1 == '>') {
            this.prefetchReader.read();
            ch1 = this.prefetchReader.get();
            if (ch1 == '=') {
              this.prefetchReader.read();
              operator = JavaToken.OPERATOR_UNSIGNED_RIGHT_SHIFT_ASSIGNMENT;
            }
            else
              operator = JavaToken.OPERATOR_UNSIGNED_RIGHT_SHIFT;
          }
          else
            operator = JavaToken.OPERATOR_RIGHT_SHIFT;
        }
        else
          operator = JavaToken.OPERATOR_GREATER;
        break;
      case '&':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_AND_ASSIGNMENT;
        }
        else if (ch1 == '&') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_OPTIONAL_AND;
        }
        else
          operator = JavaToken.OPERATOR_REQUIRED_AND;
        break;
      case '|':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_OR_ASSIGNMENT;
        }
        else if (ch1 == '|') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_OPTIONAL_OR;
        }
        else
          operator = JavaToken.OPERATOR_REQUIRED_OR;
        break;
      case '+':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_ADD_ASSIGNMENT;
        }
        else if (ch1 == '+') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_INCREMENT;
        }
        else
          operator = JavaToken.OPERATOR_ADD;
        break;
      case '-':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_SUB_ASSIGNMENT;
        }
        else if (ch1 == '-') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_DECREMENT;
        }
        else
          operator = JavaToken.OPERATOR_SUB;
        break;
      case '*':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_MULTIPLY_ASSIGNMENT;
        }
        else
          operator = JavaToken.OPERATOR_MULTIPLY;
        break;
      case '/':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_DIVIDE_ASSIGNMENT;
        }
        else
          operator = JavaToken.OPERATOR_DIVIDE;
        break;
      case '~':
        operator = JavaToken.OPERATOR_OPPOSITE;
        break;
      case '?':
        operator = JavaToken.OPERATOR_QUESTION;
        break;
      case ':':
        operator = JavaToken.OPERATOR_COLON;
        break;
      case '^':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_XOR_ASSIGNMENT;
        }
        else
          operator = JavaToken.OPERATOR_XOR;
        break;
      case '%':
        if (ch1 == '=') {
          this.prefetchReader.read();
          operator = JavaToken.OPERATOR_MODULO_ASSIGNMENT;
        }
        else
          operator = JavaToken.OPERATOR_MODULO;
        break;
      default:
        this.syntaxError(ERROR_ILLEGAL_CHARACTER);
        return null; // Not accessible
    }
    this.currentToken.setOperator(operator);
    return this.currentToken;
  }

  public Localizer getLocalizer() {
    return (Localizer)this;
  }

  public void close() throws IOException {
    this.prefetchReader.close();
  }
}
